# sage_setup: distribution = sagemath-repl
"""
Interface to the Sage cleaner

Triva Note: For the name "sage-cleaner", think of the
"The Cleaner" from Pulp Fiction:
http://www.frankjankowski.de/quiz/illus/keitel.jpg
"""
# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import os
import atexit
import tempfile

_spd = tempfile.TemporaryDirectory()
SAGE_SPAWNED_PROCESS_FILE = os.path.join(_spd.name, "spawned_processes")
atexit.register(lambda: _spd.cleanup())


def start_cleaner():
    """
    Start ``sage-cleaner`` in a new process group.
    """
    if not os.fork():
        os.setpgid(os.getpid(), os.getpid())

        # Redirect stdin, stdout, stderr to /dev/null
        with open(os.devnull, "r+") as f:
            os.dup2(f.fileno(), 0)
            os.dup2(f.fileno(), 1)
            os.dup2(f.fileno(), 2)

        # Close all other file descriptors
        try:
            maxopenfiles = os.sysconf("SC_OPEN_MAX")
            if maxopenfiles <= 0:
                raise ValueError
        except ValueError:
            maxopenfiles = 1024
        os.closerange(3, maxopenfiles)

        os.execlp("sage-cleaner", "sage-cleaner")



# sage_setup: distribution = sagemath-repl
r"""
The ``pretty_print`` command

Works similar to the ``print`` function, except that it always tries
to use a rich output for an object, as specified via the text display
preference. If such a rich output is not available, it falls back on the
plain text.

EXAMPLES::

    sage: pretty_print(1, 2, 3)
    1 2 3

    sage: pretty_print(x^2 / (x + 1))                                                   # needs sage.symbolic
    x^2/(x + 1)

TESTS::

    sage: dm = get_display_manager()
    sage: dm.preferences.text = 'ascii_art'

EXAMPLES::

    sage: %display ascii_art  # not tested
    sage: pretty_print(x^2 / (x + 1))                                                   # needs sage.symbolic
       2
      x
    -----
    x + 1

TESTS:

After the previous example, we need to reset the text display preferences::

    sage: dm.preferences.text = None

EXAMPLES:

Printing a graphics object just prints a string, whereas
:func:`pretty_print` does not print anything and just shows the
graphics instead::

    sage: print(plot(sin))                                                              # needs sage.plot sage.symbolic
    Graphics object consisting of 1 graphics primitive
    sage: pretty_print(plot(sin))                                                       # needs sage.plot sage.symbolic
"""

# ****************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.structure.sage_object import SageObject
from sage.repl.rich_output import get_display_manager


class SequencePrettyPrinter(SageObject):

    def __init__(self, *args, **kwds):
        r"""
        Pretty Printer for Multiple Arguments.

        INPUT/OUTPUT:

        Same as :func:`pretty_print`, except that the number of
        arguments must be >= 2. Otherwise its not a sequence of things
        to print.

        EXAMPLES::

            sage: pretty_print(1, 2, 3)   # indirect doctest
            1 2 3
            sage: from sage.repl.rich_output.pretty_print import SequencePrettyPrinter
            sage: SequencePrettyPrinter(1, 2, 3).pretty_print()
            1 2 3
        """
        self.args = args
        assert len(self.args) >= 2
        self.kwds = kwds

    def is_homogeneous(self, common_type):
        """
        Return whether the pretty print items are homogeneous.

        INPUT:

        - ``common_type`` -- a type

        OUTPUT:

        boolean; whether all items to be pretty printed are of said
        type.

        EXAMPLES::

            sage: from sage.repl.rich_output.pretty_print import SequencePrettyPrinter
            sage: seq = SequencePrettyPrinter(1, 2, 3)
            sage: seq.is_homogeneous(Integer)
            True
            sage: seq.is_homogeneous(str)
            False
        """
        return all(isinstance(arg, common_type) for arg in self.args)

    def _concatenate_graphs(self):
        """
        Plot multiple graphs into a single plot.

        OUTPUT: a graphics object

        EXAMPLES::

            sage: from sage.repl.rich_output.pretty_print import SequencePrettyPrinter
            sage: plt = SequencePrettyPrinter(*list(graphs(3)))._concatenate_graphs()   # needs sage.graphs sage.plot
            sage: type(plt)                                                             # needs sage.graphs sage.plot
            <class 'sage.plot.multigraphics.GraphicsArray'>
            sage: plt                                                                   # needs sage.graphs sage.plot
            Graphics Array of size 1 x 4
        """
        import sage.graphs.graph_list as graphs_list
        return graphs_list.to_graphics_array(self.args, **self.kwds)

    def _concatenate_graphics(self):
        """
        Combine multiple graphics objects into one graphics array.

        OUTPUT: a graphics array

        EXAMPLES::

            sage: from sage.repl.rich_output.pretty_print import SequencePrettyPrinter
            sage: ga = SequencePrettyPrinter(*[Graphics()]*5)._concatenate_graphics()   # needs sage.plot
            sage: type(ga)                                                              # needs sage.plot
            <class 'sage.plot.multigraphics.GraphicsArray'>
            sage: ga.nrows(), ga.ncols()                                                # needs sage.plot
            (2, 4)
        """
        from sage.plot.plot import graphics_array
        return graphics_array(self.args, ncols=4, **self.kwds)

    def pretty_print(self):
        """
        Actually do the pretty print.

        EXAMPLES::

            sage: from sage.repl.rich_output.pretty_print import SequencePrettyPrinter
            sage: SequencePrettyPrinter(1, 2, 3).pretty_print()
            1 2 3

        The keyword arguments are only used the first time graphics
        output is generated::

            sage: seq = SequencePrettyPrinter(Graph(), Graph(), edge_labels=True)       # needs sage.graphs sage.plot
            sage: seq.pretty_print()   # does not pass edge_labels to graphics object   # needs sage.graphs sage.plot
            sage: seq._concatenate_graphs().show(edge_labels=True)                      # needs sage.graphs sage.plot
            Traceback (most recent call last):
            ...
            TypeError: ...matplotlib() got an unexpected keyword argument 'edge_labels'...
        """
        try:
            from sage.plot.plot import Graphics
        except ImportError:
            Graphics = ()
        try:
            from sage.graphs.graph import GenericGraph
        except ImportError:
            GenericGraph = ()
        if self.is_homogeneous(GenericGraph):
            args = self._concatenate_graphs()
            kwds = dict()
        elif self.is_homogeneous(Graphics):
            args = self._concatenate_graphics()
            kwds = dict()
        else:
            args = self.args
            kwds = dict(self.kwds)
            kwds['concatenate'] = True
        get_display_manager().display_immediately(args, **kwds)


def pretty_print(*args, **kwds):
    r"""
    Pretty print the arguments using rich output if available.

    This function is similar to ``print()``, except that a rich output
    representation such as ``ascii_art`` or Latex is printed instead of the
    string representation.

    Note that the output depends on the global display preferences specified
    via
    :meth:`~sage.repl.rich_output.display_manager.DisplayManager.preferences`.
    If the display preference for ``text`` is not specified, Latex output is
    preferred.

    For graphical objects, a graphical output is used.

    For certain homogeneous multiple positional arguments, a suitable
    combined graphical output is generated. In particular, graphs and
    plots are treated special. Otherwise this function concatenates the
    textual representations.

    INPUT:

    - ``*args`` -- any number of positional arguments; the objects to
      pretty print

    - ``**kwds`` -- optional keyword arguments that are passed to the
      rich representation. Examples include:

        - ``dpi`` -- dots per inch

        - ``figsize``- [width, height] (same for square aspect)

        - ``axes`` -- (default: ``True``)

        - ``fontsize`` -- positive integer

        - ``frame`` -- boolean (default: ``False``); draw a MATLAB-like frame around
          the image

    EXAMPLES::

        sage: pretty_print(ZZ)
        Integer Ring

        sage: pretty_print("Integers = ", ZZ) # trac 11775
        'Integers = ' Integer Ring

    To typeset LaTeX code as-is, use :class:`LatexExpr`::

        sage: pretty_print(LatexExpr(r"\frac{x^2 + 1}{x - 2}"))
        \frac{x^2 + 1}{x - 2}

    For text-based backends, the default text display preference is to output
    plain text which is usually the same as using ``print()``::

        sage: pretty_print(x^2 / (x + 1))                                               # needs sage.symbolic
        x^2/(x + 1)

        sage: t = BinaryTrees(3).first()                                                # needs sage.graphs
        sage: pretty_print(t)                                                           # needs sage.graphs
        [., [., [., .]]]
        sage: print(t)                                                                  # needs sage.graphs
        [., [., [., .]]]

    TESTS::

        sage: dm = get_display_manager()
        sage: dm.preferences.text = 'ascii_art'

    EXAMPLES:

    Changing the text display preference affects the output of this function.
    The following illustrates a possible use-case::

        sage: %display ascii_art  # not tested
        sage: for t in BinaryTrees(3)[:3]:                                              # needs sage.graphs
        ....:     pretty_print(t)
        o
         \
          o
           \
            o
        o
         \
          o
         /
        o
          o
         / \
        o   o

        sage: pretty_print(x^2 / (x + 1))                                               # needs sage.symbolic
           2
          x
        -----
        x + 1

    TESTS:

    After the previous example, we need to reset the text display preferences::

        sage: dm.preferences.text = None

    ::

        sage: # needs sage.plot sage.symbolic
        sage: plt = plot(sin)
        sage: pretty_print(plt)             # graphics output
        sage: pretty_print(plt, plt)        # graphics output
        sage: pretty_print(ZZ, 123, plt)
        Integer Ring 123 Graphics object consisting of 1 graphics primitive
    """
    dm = get_display_manager()
    old_preferences_text = dm.preferences.text
    try:
        if dm.preferences.text is None:
            dm.preferences.text = 'latex'
        if len(args) == 0:
            pass
        elif len(args) == 1:
            dm.display_immediately(*args, **kwds)
        else:
            SequencePrettyPrinter(*args, **kwds).pretty_print()
    finally:
        dm.preferences.text = old_preferences_text


def show(*args, **kwds):
    r"""
    Alias for ``pretty_print``.

    This function is an alias for :func:`pretty_print`.

    INPUT/OUTPUT:

    See :func:`pretty_print`. Except if the argument is a graph, in
    which case it is plotted instead.

    EXAMPLES::

        sage: show(1)
        1
    """
    try:
        from sage.graphs.generic_graph import GenericGraph
    except ImportError:
        GenericGraph = ()
    if len(args) == 1 and isinstance(args[0], GenericGraph):
        # Graphs are special, they ride the short bus...
        # Please, somebody help me get rid of this! #18289
        args[0].show()
        return
    pretty_print(*args, **kwds)



# sage_setup: distribution = sagemath-repl
"""
Quitting interfaces
"""

################################################################################
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of (any version of) the GNU
#  General Public License (GPL). The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
################################################################################

from __future__ import annotations

import os
import subprocess
import sys
from typing import TYPE_CHECKING

from sage.env import DOT_SAGE, HOSTNAME
from sage.misc.cachefunc import cached_function

if TYPE_CHECKING:
    from weakref import ReferenceType

    from sage.interfaces.expect import Expect


@cached_function
def sage_spawned_process_file() -> str:
    """
    EXAMPLES::

        sage: from sage.interfaces.quit import sage_spawned_process_file
        sage: len(sage_spawned_process_file()) > 1
        True
    """
    # This is the old value of SAGE_TMP. Until sage-cleaner is
    # completely removed, we need to leave these spawned_processes
    # files where sage-cleaner will look for them.
    d = os.path.join(DOT_SAGE, "temp", HOSTNAME, str(os.getpid()))
    os.makedirs(d, exist_ok=True)
    return os.path.join(d, "spawned_processes")


def register_spawned_process(pid: int, cmd: str = "") -> None:
    """
    Write a line to the ``spawned_processes`` file with the given
    ``pid`` and ``cmd``.
    """
    if cmd != "":
        cmd = cmd.strip().split()[0]
    # This is safe, since only this process writes to this file.
    try:
        with open(sage_spawned_process_file(), "a") as file:
            file.write("%s %s\n" % (pid, cmd))
    except OSError:
        pass
    else:
        # If sage is being used as a python library, we need to launch
        # the cleaner ourselves upon being told that there will be
        # something to clean.
        from sage.interfaces.cleaner import start_cleaner

        start_cleaner()


expect_objects: list[ReferenceType[Expect]] = []


def expect_quitall(verbose: bool = False) -> None:
    """
    EXAMPLES::

        sage: sage.interfaces.quit.expect_quitall()
        sage: gp.eval('a=10')                                                           # needs sage.libs.pari
        '10'
        sage: gp('a')                                                                   # needs sage.libs.pari
        10
        sage: sage.interfaces.quit.expect_quitall()
        sage: gp('a')                                                                   # needs sage.libs.pari
        a
        sage: sage.interfaces.quit.expect_quitall(verbose=True)                         # needs sage.libs.pari
        Exiting PARI/GP interpreter with PID ... running .../gp --fast --emacs --quiet --stacksize 10000000
    """
    for reference in expect_objects:
        process = reference()
        if process is not None:
            try:
                process.quit(verbose=verbose)
            except RuntimeError:
                pass
    kill_spawned_jobs()


def kill_spawned_jobs(verbose: bool = False):
    """
    INPUT:

    - ``verbose`` -- boolean (default: ``False``); if ``True``, display a
      message each time a process is sent a kill signal

    EXAMPLES::

        sage: gp.eval('a=10')                                                           # needs sage.libs.pari
        '10'
        sage: sage.interfaces.quit.kill_spawned_jobs(verbose=False)
        sage: sage.interfaces.quit.expect_quitall()
        sage: gp.eval('a=10')                                                           # needs sage.libs.pari
        '10'
        sage: sage.interfaces.quit.kill_spawned_jobs(verbose=True)                      # needs sage.libs.pari
        Killing spawned job ...

    After doing the above, we do the following to avoid confusion in other doctests::

        sage: sage.interfaces.quit.expect_quitall()
    """
    fname = sage_spawned_process_file()
    if not os.path.exists(fname):
        return

    with open(fname) as file:
        for line in file:
            i = line.find(" ")
            pid = line[:i].strip()
            try:
                if verbose:
                    print("Killing spawned job %s" % pid)
                if sys.platform == "win32":
                    # From https://stackoverflow.com/a/47756757/873661
                    subprocess.call(["taskkill", "/F", "/T", "/PID", pid])
                else:
                    os.killpg(int(pid), 9)
            except OSError:
                pass


def is_running(pid: int) -> bool:
    """
    Return ``True`` if and only if there is a process with id pid running.
    """
    try:
        os.kill(int(pid), 0)
        return True
    except (OSError, ValueError):
        return False


def invalidate_all() -> None:
    """
    Invalidate all of the expect interfaces.

    This is used, e.g., by the fork-based ``@parallel`` decorator.

    EXAMPLES::

        sage: # needs sage.libs.pari sage.symbolic
        sage: a = maxima(2); b = gp(3)
        sage: a, b
        (2, 3)
        sage: sage.interfaces.quit.invalidate_all()
        sage: a
        (invalid Maxima object -- The maxima session in which this object was defined is no longer running.)
        sage: b
        (invalid PARI/GP interpreter object -- The pari session in which this object was defined is no longer running.)

    However the maxima and gp sessions should still work out, though with their state reset::

        sage: a = maxima(2); b = gp(3)                                                  # needs sage.libs.pari sage.symbolic
        sage: a, b                                                                      # needs sage.libs.pari sage.symbolic
        (2, 3)
    """
    for reference in expect_objects:
        process = reference()
        if process:
            process.detach()



# sage_setup: distribution = sagemath-repl
r"""
Display Preferences

This class is used to express display preferences that are not simply
a choice of a particular output format. For example, whether to prefer
vector over raster graphics. By convention, the value ``None`` is
always a valid value for a preference and means no particular
preference.

EXAMPLES::

    sage: from sage.repl.rich_output.preferences import DisplayPreferences
    sage: prefs = DisplayPreferences()
    sage: prefs.available_options()
    (align_latex, graphics, supplemental_plot, text)
    sage: prefs.text is None
    True
    sage: prefs.text = 'ascii_art'
    sage: prefs.text
    'ascii_art'
    sage: prefs
    Display preferences:
    * align_latex is not specified
    * graphics is not specified
    * supplemental_plot is not specified
    * text = ascii_art

Properties can be unset by deleting them or by assigning ``None``::

    sage: prefs.text = 'ascii_art'
    sage: del prefs.text
    sage: prefs.text is None
    True

    sage: prefs.text = 'ascii_art'
    sage: prefs.text = None
    sage: prefs.text is None
    True

Properties have documentation attached::

    sage: import pydoc
    sage: doc = pydoc.render_doc(prefs)
    sage: assert ' graphics' in doc
    sage: assert '     Preferred graphics format' in doc
    sage: assert ' text' in doc
    sage: assert '     Which textual representation is preferred' in doc

Values can also be specified as keyword arguments to the constructor::

    sage: DisplayPreferences(text='latex')
    Display preferences:
    * align_latex is not specified
    * graphics is not specified
    * supplemental_plot is not specified
    * text = latex

.. TODO::

    A value-checking preference system should be used elsewhere in
    Sage, too. The class here is just a simple implementation, a
    proper implementation would use a metaclass to construct the
    preference items.
"""

# ****************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from textwrap import dedent

from sage.structure.sage_object import SageObject


class Property(property):

    def __init__(self, name, allowed_values, doc=None):
        r"""
        Preference item.

        INPUT:

        - ``name`` -- string; the name of the property

        - ``allowed_values`` -- list/tuple/iterable of allowed values

        - ``doc`` -- string (optional); the docstring of the property

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import Property
            sage: prop = Property('foo', [0, 1, 2], 'The Foo Property')
            sage: prop.__doc__
            'The Foo Property\n\nAllowed values:\n\n* ``None`` (default): no preference\n\n* 0\n\n* 1\n\n* 2'
            sage: prop.allowed_values
            (0, 1, 2)
        """
        self.name = name
        self.underscore_name = '_{0}'.format(name)
        self.allowed_values = tuple(allowed_values)
        self.__doc__ = doc = self._make_doc(doc)
        super().__init__(fget=self.getter, fset=self.setter,
                         fdel=self.deleter, doc=doc)

    def _make_doc(self, doc):
        """
        Generate the documentation.

        INPUT:

        - ``doc`` -- the title line of the documentation

        OUTPUT:

        string; the docstring with auto-generated documentation about
        the allowed values added

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import Property
            sage: prop = Property('foo', [0, 1, 2], 'The Foo Property')
            sage: print(prop._make_doc('this is the title'))
            this is the title
            <BLANKLINE>
            Allowed values:
            <BLANKLINE>
            * ``None`` (default): no preference
            <BLANKLINE>
            * 0
            <BLANKLINE>
            * 1
            <BLANKLINE>
            * 2
        """
        doc = dedent(doc)
        doc += '\n\n'
        doc += 'Allowed values:\n\n'
        values_doc = ['* ``None`` (default): no preference']
        values_doc.extend('* {0}'.format(repr(value))
                          for value in self.allowed_values)
        return doc + '\n\n'.join(values_doc)

    def __repr__(self):
        """
        Return a string representation.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import Property
            sage: prop = Property('foo', [0, 1, 2], 'The Foo Property')
            sage: prop.__repr__()
            'foo'
        """
        return self.name

    def getter(self, prefs):
        """
        Get the current value of the property.

        INPUT:

        - ``prefs`` -- the :class:`PreferencesABC` instance that the
          property is bound to

        OUTPUT: one of the allowed values or ``None`` if not set

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import Property, PreferencesABC
            sage: prop = Property('foo', [0, 1, 2], 'The Foo Property')
            sage: prefs = PreferencesABC()
            sage: prop.getter(prefs) is None
            True
            sage: prop.setter(prefs, 1)
            sage: prop.getter(prefs)
            1
        """
        try:
            return getattr(prefs, self.underscore_name)
        except AttributeError:
            return None

    def setter(self, prefs, value):
        """
        Get the current value of the property.

        INPUT:

        - ``prefs`` -- the :class:`PreferencesABC` instance that the
          property is bound to

        - ``value`` -- anything. The new value of the
          property. Setting a property to ``None`` is equivalent to
          deleting the value.

        OUTPUT:

        This method does not return anything. A :exc:`ValueError` is
        raised if the given ``value`` is not one of the allowed
        values.

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import Property, PreferencesABC
            sage: prop = Property('foo', [0, 1, 2], 'The Foo Property')
            sage: prefs = PreferencesABC()
            sage: prop.getter(prefs) is None
            True
            sage: prop.setter(prefs, 1)
            sage: prop.getter(prefs)
            1

            sage: prop.setter(prefs, None)
            sage: prop.getter(prefs) is None
            True
        """
        if value is None:
            return self.deleter(prefs)
        allowed = self.allowed_values
        if value not in allowed:
            raise ValueError('value must be unset (None) or one of {0}, got {1}'
                             .format(allowed, value))
        setattr(prefs, self.underscore_name, value)

    def deleter(self, prefs):
        """
        Delete the current value of the property.

        INPUT:

        - ``prefs`` -- the :class:`PreferencesABC` instance that the
          property is bound to

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import Property, PreferencesABC
            sage: prop = Property('foo', [0, 1, 2], 'The Foo Property')
            sage: prefs = PreferencesABC()
            sage: prop.getter(prefs) is None
            True
            sage: prop.setter(prefs, 1)
            sage: prop.deleter(prefs)
            sage: prop.getter(prefs) is None
            True
        """
        underscore_name = self.underscore_name
        try:
            delattr(prefs, underscore_name)
        except AttributeError:
            pass


class PreferencesABC(SageObject):

    def __init__(self, *args, **kwds):
        """
        Preferences for displaying graphics.

        These can be preferences expressed by the user or by the
        display backend. They are specified as keyword arguments.

        INPUT:

        - ``*args*`` -- positional arguments are preferences
          instances. The property values will be inherited from left
          to right, that is, later parents override values from
          earlier parents.

        - ``**kwds`` -- keyword arguments; will be used to initialize
          properties, and override inherited values if necessary

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import DisplayPreferences
            sage: p1 = DisplayPreferences(graphics='vector')
            sage: p2 = DisplayPreferences(graphics='raster')
            sage: DisplayPreferences(p1, p2)
            Display preferences:
            * align_latex is not specified
            * graphics = raster
            * supplemental_plot is not specified
            * text is not specified

        If specified in the opposite order, the setting from ``p1`` is
        inherited::

            sage: DisplayPreferences(p2, p1)
            Display preferences:
            * align_latex is not specified
            * graphics = vector
            * supplemental_plot is not specified
            * text is not specified

        Further keywords override::

            sage: DisplayPreferences(p2, p1, graphics='disable')
            Display preferences:
            * align_latex is not specified
            * graphics = disable
            * supplemental_plot is not specified
            * text is not specified
        """
        for parent in args:
            for option in self.available_options():
                value = option.getter(parent)
                if value is not None:
                    option.setter(self, value)
        for key, value in kwds.items():
            setattr(self, key, value)

    @classmethod
    def _add_option(cls, name, values, doc):
        """
        Add an option to the preferences system.

        This method should only be called during the import of
        :mod:`sage.repl.rich_output.preferences`.

        INPUT:

        - ``name`` -- the name of the option

        - ``values`` -- the allowed values

        - ``doc`` -- docstring

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import PreferencesABC
            sage: class MyPrefs(PreferencesABC):
            ....:     pass
            sage: MyPrefs._add_option('foo', [0, 1, 2], 'The Foo Option')
            sage: prefs = MyPrefs()
            sage: prefs.foo
            sage: prefs.foo = 0
            sage: prefs.foo
            0
        """
        prop = Property(name, values, doc)
        setattr(cls, name, prop)

    def available_options(self):
        """
        Return the available options.

        OUTPUT:

        tuple of the preference items as instances of :class:`Property`

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import DisplayPreferences
            sage: DisplayPreferences().available_options()
            (align_latex, graphics, supplemental_plot, text)
        """
        options = []
        for value in self.__class__.__dict__.values():
            if isinstance(value, Property):
                options.append(value)
        return tuple(sorted(options, key=str))

    def _repr_(self):
        r"""
        Return a string representation.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.repl.rich_output.preferences import DisplayPreferences
            sage: DisplayPreferences()._repr_()
            'Display preferences:\n* align_latex is not specified\n* graphics is not specified\n* supplemental_plot is not specified\n* text is not specified'
        """
        s = ['Display preferences:']
        for opt in self.available_options():
            value = opt.getter(self)
            if value is None:
                s += ['* {0} is not specified'.format(opt.name)]
            else:
                s += ['* {0} = {1}'.format(opt.name, value)]
        return '\n'.join(s)


class DisplayPreferences(PreferencesABC):
    pass


DisplayPreferences._add_option(
    'text',
    ('plain', 'ascii_art', 'unicode_art', 'latex'),
    """
    Which textual representation is preferred
    """
)


DisplayPreferences._add_option(
    'align_latex',
    ('center', 'left'),
    """
    Preferred mode of latex displays
    """
)


DisplayPreferences._add_option(
    'graphics',
    ('disable', 'vector', 'raster'),
    """
    Preferred graphics format
    """
)


DisplayPreferences._add_option(
    'supplemental_plot',
    ('always', 'never'),
    """
    Whether to graphically display graphs and other graph-like objects
    that implement rich output. When not specified small objects are
    show graphically and large objects as textual overview.
    """
)



# sage_setup: distribution = sagemath-repl
# sage.doctest: needs sage.modules
r"""
User-interface globals

These represent "globals" in the top-level user interface (command
line, notebook). When starting up Sage, ``user_globals`` becomes
identical to the ``globals()`` of the top-level frame. Later on,
functions can inject globals.

.. NOTE::

    Injecting globals is a feature which has some legitimate uses, but
    beware of over-using it. A natural use case for injecting globals is
    for "compile-and-import" functions, such as :func:`cython`.

EXAMPLES:

This is how a typical user interface initializes the globals::

    sage: ui_globals = globals()  # or wherever the user interface stores its globals
    sage: from sage import all_cmdline
    sage: from sage.repl.user_globals import initialize_globals
    sage: _ = initialize_globals(all_cmdline, ui_globals)

Now everything which was imported in ``all_cmdline`` is available as a
global::

    sage: from sage.repl.user_globals import get_global, set_global
    sage: get_global("Matrix")
    <cyfunction matrix at ...>

This is exactly the same::

    sage: ui_globals["Matrix"]
    <cyfunction matrix at ...>

We inject a global::

    sage: set_global("myvar", "Hello World!")
    sage: get_global("myvar")
    'Hello World!'

If we set up things correctly, this new variable is now actually
available as global::

    sage: myvar
    'Hello World!'

AUTHORS:

- Jeroen Demeyer (2015-03-30): initial version (:issue:`12446`)
"""

# ****************************************************************************
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


user_globals = None


def _check():
    """
    Raise :exc:`RuntimeError` if ``user_globals`` has not been initialized.

    EXAMPLES::

        sage: import sage.repl.user_globals
        sage: sage.repl.user_globals._check()
        sage: sage.repl.user_globals.user_globals = None
        sage: sage.repl.user_globals._check()
        Traceback (most recent call last):
        ...
        RuntimeError: the user-space globals dictionary has not been initialized...
    """
    if user_globals is None:
        raise RuntimeError(
            "the user-space globals dictionary has not been initialized. "
            "Use initialize_globals() or set_globals() or use a different "
            "function which doesn't need these globals")


def get_globals():
    """
    Return the dictionary of all user globals.

    EXAMPLES::

        sage: from sage.repl.user_globals import get_globals, initialize_globals
        sage: initialize_globals(sage.all)
        sage: get_globals()["Matrix"]
        <cyfunction matrix at ...>
    """
    _check()
    return user_globals


def set_globals(g: dict) -> None:
    """
    Set the dictionary of all user globals to ``g``.

    INPUT:

    - ``g`` -- dictionary; typically, this will be some dictionary
      given by the user interface or just ``globals()``

    EXAMPLES::

        sage: from sage.repl.user_globals import get_globals, set_globals
        sage: my_dict = dict()
        sage: set_globals(my_dict)
        sage: my_dict is get_globals()
        True
    """
    global user_globals
    user_globals = g


def initialize_globals(all, g=None):
    """
    Set the user globals dictionary to ``g`` and assign everything
    which was imported in module ``all`` as global.

    INPUT:

    - ``all`` -- a module whose globals will be injected

    - ``g`` -- dictionary; see :func:`set_globals`. If this is
      ``None``, keep the current globals dictionary.

    EXAMPLES::

        sage: my_globs = {"foo": "bar"}
        sage: from sage.repl.user_globals import initialize_globals
        sage: initialize_globals(sage.all, my_globs)
        sage: my_globs["foo"]
        'bar'
        sage: my_globs["Matrix"]
        <cyfunction matrix at ...>

    Remove ``Matrix`` from the globals and initialize again without
    changing the dictionary::

        sage: del my_globs["Matrix"]
        sage: initialize_globals(sage.all)
        sage: my_globs["Matrix"]
        <cyfunction matrix at ...>
    """
    if g is not None:
        set_globals(g)
    for key in dir(all):
        if key[0] != '_':
            user_globals[key] = getattr(all, key)
    from sage.misc.lazy_import import clean_namespace
    clean_namespace(user_globals)


def get_global(name):
    """
    Return the value of global variable ``name``.

    Raise :exc:`NameError` if there is no such global variable.

    INPUT:

    - ``name`` -- string representing a variable name

    OUTPUT: the value of variable ``name``

    EXAMPLES::

        sage: from sage.repl.user_globals import get_global
        sage: the_answer = 42
        sage: get_global("the_answer")
        42
        sage: get_global("the_question")
        Traceback (most recent call last):
        ...
        NameError: name 'the_question' is not defined
    """
    _check()
    try:
        return user_globals[name]
    except KeyError:
        raise NameError("name {!r} is not defined".format(name))


def set_global(name, value):
    """
    Assign ``value`` to global variable ``name``. This is equivalent
    to executing ``name = value`` in the global namespace.

    INPUT:

    - ``name`` -- string representing a variable name

    - ``value`` -- a value to assign to the variable

    EXAMPLES::

        sage: from sage.repl.user_globals import set_global
        sage: set_global("the_answer", 42)
        sage: the_answer
        42
    """
    _check()
    user_globals[name] = value
